#! /usr/local/bin/perl -w

use Text::Wrap;
use Text::Tabs;

# chapters: we get run once per chapter.
#  guide1, the chapter called Palm OS User Interface Functions (pp 193-394)
#   Category Functions
#   ClipBoard Functions
#   Control Functions
#   Field UI Functions
#   Font Functions
#   Form Functions
#   Character Attribute Functions
#   Graffiti Manager Functions
#   GraffitiShift Functions
#   Insertion Point Functions
#   Key Manager Functions
#   List UI Functions
#   Menu Functions
#   Miscellaneous User Interface Functions
#   Pen Manager Functions
#   Scrollbar Functions
#   Functions for system Use Only
#   Time Selection Functions
#   Table Functions
#   Window Functions
#  guide2, chapter called Palm OS System Functions (pp 45-145)
#   Alarm Manager API
#   Error Manager Functions
#   Event Manager Functions
#   Functions for System Use Only
#   Feature Functions
#   For System Use Only
#   Find Functions
#   Float Manager Functions
#   Miscellaneous System Functions
#   System Preferences Functions
#   Password Functions
#   String Manager Functions
#   Sound Manager Functions
#   System Functions
#   Time Manager Functions
#  guide3, "Memory Management Functions", pp 47-135
#   Memory Manager Functions
#   Data and Resource Manager Functions
#  guide3, "Communications Functions", pp 155-180
#   Serial Manager Functions
#   Serial Link Manager Functions
#   Miscellaneous Communications Functions
#  guide3, "Net Library Functions", pp 198-252
#  (here it gets a bit weird)
#   Configuration (pp 253-271), there are a few tables
#   (Supported Socket Functions et all is a series of tables. They describe
#    functions that are available but not listed elsewhere. Maybe. Not enough
#    to worry about now)

# each function definition looks nominally like this:

#    functionname
#  Purpose   stuff
#  Prototype  void functionname\s*(type  name,
#        type name,
#        type name)
#
#  Parameters  name   stuff
#     name stuff
#         more stuff
#     name
#         stuff
#  Result  stuff
#  See Also  stuff
#        stuff
#
#      functionname
#    (repeats)

# other header names:
#   Caveat
#   comment

my $state = "";
# one of ("" qw(infunction Purpose Prototype Parameters Result SeeAlso 
#         Caveat comment))

my (@functions);  # the functions being created. Each is a hashref
#  {  name => function_name,
#     purpose => purpose text,
#     prototype => [returntype, {type, name, dir, text}, {type, name, stuff}.],
#     result => result text
#     seealso => text
#     caveat => text
#     comment => text
#  }

my %current;  # the hash for the function currently being parsed

my $debug_empty_parms = 0;

sub nextline {
    ($_) = @_;

#    print "$state: $_";

    if (!$state) {
	if (/^\s*(\w+)\s*$/) {
	    # function names tend to appear by themselves
	    $function_name = $1;
	    $state = "infunction";
	    return;
	}
    }

    # single words might be a function name, or might be more text.
    if (/^\s*(\w+)\s*$/) {
	$possible_function_name = $1;
    }

    # word word word Functions might be a chapter title
    if (/\bFunctions\s*$/ or /\bManager API\s*$/) {
	my $copy = $_;
	$copy =~ s/^\s+//; $copy =~ s/\s+$//;
	my(@words) = split(/\s+/,$copy);
	if (@words <= 4) {
	    $possible_chapter_name = join(' ',@words);
	} else {
	}
    }

    if (/^\s*Purpose/) {
	# start of a purpose section. Remember the function name. Finish off
	# the old function.
	if (%current and !defined($current{'purpose'})) {
	    # the last function didn't have a purpose. Add one.
	    $current{'purpose'} = "<no purpose>";
	}
	# the function name (as recorded in $possible_function_name) probably
	# got included in the last section. Remove it.
	my $lastref;
	$lastref = \$current{'result'} if ($state eq 'Result');
	$lastref = \$current{'seealso'} if ($state eq 'SeeAlso');
	$lastref = \$current{'caveat'} if ($state eq 'Caveat');
	$lastref = \$current{'comment'} if ($state eq 'comment');
	if ($state ne 'infunction' and $state ne 'Prototype') {
	    # infunction is the first function: there is no preceeding text
	    # to fix. Prototype probably follows a "for system use only"
	    # function that only has a prototype section.
	    if (!$lastref) {
		print STDERR "no lastref, state is $state, line $. in $current{'name'}\n";
	    } else {
		$$lastref =~ s/\Q$possible_function_name\E\s*$//;
		$$lastref =~ s/\Q$possible_chapter_name\E\s*$//
		  if $possible_chapter_name;
	    }
	}

	&finish_off_function;
	$current{'name'} = $possible_function_name || $function_name;
	if ($possible_chapter_name) {
	    $current_chapter_name = $possible_chapter_name;
	    $possible_chapter_name = "";
	}
	s/^\s*Purpose\s*//;
	$current{'purpose'} .= $_;
	$state = 'Purpose';
	return;
    }

    if (/^\s*Prototype/) {
	# prototype section. Getting the parm names is important because
	# it helps us parse the Parameters section
	$state = 'Prototype';

	if ($current{'prototype_text'}) {
	    # double prototype. This function didn't have a purpose line.
	    # fake one
	    &finish_off_function; # clears %current
	    $current{'purpose'} = "<no purpose>";
	    $current{'name'} = $possible_function_name || $function_name;
	}

	s/^\s*Prototype\s*//;
	$current{'prototype_text'} .= $_;
	if (/\)/) {
	    # the prototype started and ended on the same line.
	    &parse_prototype();
	}
	return;
    }

    if (/^\s*Parameters/) {
	unless ($current{'prototype'}) {
	    die "got Parameters without first seeing Prototype in $function_name, line $.";
	}
	$state = 'Parameters';
	s/^\s*Parameters\s*//;
	$current{'parameters_text'} .= $_;
	return;
    }

    if (/^\s*Result/) {
	$state = 'Result';
	s/^\s*Result\s*//;
	$current{'result'} .= $_;
	return;
    }

    if (/^\s*See\s+Also/) {
	$state = 'SeeAlso';
	s/^\s*See\s+Also\s*//;
	$current{'seealso'} .= $_;
	return;
    }

    if (/^\s*Caveats?/) {
	$state = 'Caveat';
	s/^\s*Caveats?\s*//;
	$current{'caveat'} .= $_;
	return;
    }

    if (/^\s*[Cc]omments?/) {
	$state = 'comment';
	s/^\s*[Cc]omments?\s*//;
	$current{'comment'} .= $_;
	return;
    }
    if (/^\s*See Instead/) {
	$state = 'seeinstead';
	s/^\s*See Instead//;
	$current{'seeinstead'} .= $_;
	return;
    }


    if ($state eq 'Purpose') {
	$current{'purpose'} .= $_;
	return;
    }
    if ($state eq 'Prototype') {
	$current{'prototype_text'} .= $_;
	if (/\)/) {
	    &parse_prototype();
	}
	return;
    }
    if ($state eq 'Parameters') {
	$current{'parameters_text'} .= $_;
	return;
    }
    if ($state eq 'Result') {
	$current{'result'} .= $_;
	return;
    }
    if ($state eq 'SeeAlso') {
	$current{'seealso'} .= $_;
	return;
    }
    if ($state eq 'Caveat') {
	$current{'caveat'} .= $_;
	return;
    }
    if ($state eq 'comment') {
	$current{'comment'} .= $_;
	return;
    }
    if ($state eq 'seeinstead') {
	$current{'seeinstead'} .= $_;
	return;
    }

}

sub parse_prototype {
    # take the prototype text in $current{'prototype_text'} and parse it
    my $text = $current{'prototype_text'};
    $text =~ s/\n//g; # make into one line
    my($return, $name);
    if ($text =~ /\*[^\(]*\(/) {
	# there is a pointer here. bind it leftwards.
	($return, $name) = ($text =~ /^([^\*]+\*+)\s*(\w+)/);
#    } elsif ($text =~ /enum/) {
#	($return, $name) = ($text =~ /^(enum\s+\w+)\s+(\w+)$/);
    } else {
	($return, $name) = ($text =~ /^(\w+)\s+(\w+)\s*\(/);
    }
    die "prototype: no return+name in '$text', line $." unless defined($name);
    $current{'prototype'} = [$return];
    $text =~ s/^[^\(]+\(//;
    $text =~ s/\)[^\)]*$//;
    if ($text =~ /^\s*void\s*$/) {
	# no input args. prototype will just be [return_type]
	return;
    }
    my(@parms) = split(/,/, $text);
    # todo: if there are both 'x' and 'y' arguments, then the description will
    # probably be merged. Be on the lookout for these, and don't include the
    # ',y' in the description of 'x'. Insert a 'see "x" above' in the 
    # description for 'y' if there isn't any other text found
    # ditto with 'extentX', 'extentY'

    foreach my $parm (@parms) {
	$parm =~ s/^\s*//;
	$parm =~ s/\s*$//;
	my($type,$name, $prefix, $suffix);
	if ($parm =~ /^const\s+/) {
	    $prefix = 'const';
	    $parm =~ s/^const\s+//;
	}
	# Some of the parameters are arrays, they look like:
	#                              Word maxEntries,  
	#                              Char hostNames[][netDNSMaxDomainName+1],
	#                              Word priorities[], 
	# we want to pull off the array suffixes and tack them on to the
	# type suffix later. Rely on greedy matching to get them all
	if ($parm =~ s/(\[.*\])$//) {
	    $suffix = $1;
	}
	$parm =~ s/\[\]//;
	if ($parm eq '...') {
	    ($type,$name) = ('varargs','...');
	} elsif ($parm =~ /\*/) {
	    ($type,$name) = ($parm =~ /^([^\*]+\*+)\s*(\w+)/);
	} elsif ($parm =~ /enum/) {
	    ($type,$name) = ($parm =~ /^(enum\s+\w+)\s+(\w+)$/);
	} else {
	    ($type,$name) = ($parm =~ /^(\w+)\s+(\w+)$/);
	}
	if ($prefix) {
	    $type = "$prefix $type";
	}
	die "prototype parm: no type+name in '$parm', line $." 
	  unless defined($name);
	my $piece = {'type' => $type, 'name' => $name };
	$piece->{'suffix'} = $suffix if $suffix;
	push(@{$current{'prototype'}}, $piece);
    }
}

sub finish_off_function {
    # parse parameters, fill all text
    if ($current{'parameters_text'} and @{$current{'prototype'}} > 1) {
	my $protos = $current{'prototype'};
	my(@parms) = split(/\n/, $current{'parameters_text'});
	my $stuff;
	my $current_parm_num = -1;
	my $current_parm_name = "";
	my $current_parm;
	my $next_parm_num = $current_parm_num + 1;
	my $next_parm_name = $protos->[$next_parm_num+1]->{'name'};
	foreach (@parms) {
	    s/^\s*//; s/\s*$//;
#	    my $debug = $current{'name'} eq 'NetLibSocketBind';
	    print STDERR "PARM: '$_'\n" if $debug;
	    if ($next_parm_name and 
		/^(\<\-|\<\-\>|\-\>)?\s*\Q$next_parm_name\E/i) 
	    {
		# we've switched to the next parm. Update the pointers
		if (/^(\<\-|\<\-\>|\-\>)/) {
		    # a silly <- or <-> or -> direction indicator.
		    my $dir = $1;
		    $protos->[$next_parm_num+1]->{'dir'} = $dir;
		    s/^\Q$dir\E\s*//;
		    print STDERR "DIR: '$next_parm_name' '$_'\n" if $debug;
		}		
		$current_parm_num = $next_parm_num;
		$current_parm_name = $next_parm_name;
		$next_parm_num = $current_parm_num + 1;
		if ($next_parm_num < @{$protos} - 1) {
		    $next_parm_name = $protos->[$next_parm_num+1]->{'name'};
		} else {
		    $next_parm_name = ""; # there is no next
		}
		$current_parm = $protos->[$current_parm_num+1];
		# strip the parm name
		s/^\Q$current_parm_name\E\s*//i;
		if ($debug and $protos->[$next_parm_num]->{'dir'}) {
		    print STDERR "BUR: '$_'\n";
		}
	    }
	    die "no current parm, line $., function $current{'name'}, parm '$_'" unless($current_parm);
	    $current_parm->{'text'} .= "$_ "; # lost a \n in the split() above
	}
	foreach my $parmnum (1 .. $#{$protos}) {
	    if ($debug_empty_parms and 
		!defined($protos->[$parmnum]->{'text'})) {
		print STDERR "warning, no text for $current{'name'}:$protos->[$parmnum]->{'name'}\n";
	    }
	    $protos->[$parmnum]->{'text'} = 
	      compact($protos->[$parmnum]->{'text'});
	}
    }
    if ($current_chapter_name) {
	$current{'chapter_name'} = $current_chapter_name;
    }

    $current{'purpose'} = &compact($current{'purpose'});
    $current{'result'} = &compact($current{'result'});
    $current{'seealso'} = &compact($current{'seealso'});
    $current{'comment'} = &compact($current{'comment'});
    $current{'caveat'} = &compact($current{'caveat'});

    if (%current) {
	my(%copy) = %current;
	push(@functions, \%copy);
	push(@{$chapters{$current_chapter_name}}, \%copy)
	  if $current_chapter_name;
#	&dumpfunc(\%current);
    }  # else, this was the first (empty) function
    %current = ();
}


#$chapter = "Palm OS User Interface Functions";
$linenum = 0;

if ($ARGV[0] eq '--testing') {
    $testing = 1;
    shift(@ARGV);
}

while(<>) {
    $linenum++;
    # the 2nd line says "Palm OS User Interface Functions"
    # the 3rd says "Category Functions"
    # the next-to-last says "Developing Palm OS 2.0 Applications, Part I  <pp>
    #  (but the number of lines per page varies)

    # big-text headers tend to get slammed to the left margin, but smaller
    # headers can end up there too if there's a lot of text elsewhere on the
    # line. This is *not* a markup language.

    if ($linenum <= 3 and (/Functions\s*$/ or /Manager API\s*$/)) {
	# set the possible chapter name based on what we see here
	my $copy = $_;
	$copy =~ s/^\s+//; $copy =~ s/\s+$//;
	my(@words) = split(/\s+/,$copy);
	if (@words <= 4) {
	    $possible_chapter_name = join(' ',@words);
	}
    }

    next if ($linenum <= 3);

#    next if /^\d+\s*$/ or /^\s*\d+$/;  # page numbers
    if (/^\d*\s*Developing Palm OS 2\.0 Applications/) {
#	print "footer on line $linenum\n";
	next;
    }
    if (/^\cL$/) {
	$linenum = 0;
	next;
    }

    s/\245/*/; # bullet

    &nextline($_);

#    print "$linenum: $_";
#    print;
}
&finish_off_function;

sub dumpfunc {
    my($f) = @_;
    print "======================================================================\n";
    print "Name: $f->{'name'}\n";
    print &fill($f->{'purpose'}, 60, 
		"Purpose: ",
		"       : "), "\n";
    print "\n";
    foreach my $key (sort(keys(%$f))) {
	next if $key eq 'name' or $key eq 'purpose';
	if ($key eq 'prototype') {
	    my $p = $f->{$key};
	    print " return_type: ",$p->[0],"\n";
	    foreach my $i (1 .. (@$p - 1)) {
		my $a = $p->[$i];
		print "  arg$i: type: ",$a->{'type'},"\n";
		print "  arg$i: name: ",$a->{'name'},"\n";
		print "  arg$i: desc: ",&fill($a->{'text'}, 
					       60, 
					       "", 
					       "            : "),"\n";
		print "  arg$i:  dir: ",$a->{'dir'},"\n" if $a->{'dir'};
		print "  arg$i:suffix: ",$a->{'suffix'},"\n" if $a->{'suffix'};
	    }
	    next;
	}
#	next if ($key eq 'parameters_text' or $key eq 'prototype_text');
	if ($key eq 'purpose' or $key eq 'result' or $key eq 'seealso'
	    or $key eq 'comment' or $key eq 'caveat') {
	    print &fill($f->{$key}, 70, "$key: ", ' ' x length($key) . ": ");
	    print "\n";
	    next;
	}
	print "$key: $f->{$key}\n";
    }
}

# fill
sub fill {
    my($text, $cols, $firsttab, $othertabs) = @_;
    $Text::Wrap::columns = $cols;
    $text = "<notext>" unless defined($text);
    $text =~ s/(\w)-\n\s*(\w)/$1$2/g;
    $text =~ s/\n/ /g;
    $text =~ s/\s+/ /g;
    $text =~ s/^\s+//;
    $text =~ s/\s+$//;
    return wrap($firsttab, $othertabs, $text);
}

sub compact {
    my($text) = @_;
    return undef unless defined($text);
    $text =~ s/(\w)-\n\s*(\w)/$1$2/g;
    $text =~ s/\n/ /g;
    $text =~ s/\s+/ /g;
    $text =~ s/^\s+//;
    $text =~ s/\s+$//;
    return $text;
}

sub chapter_to_nodename {
    my($chapter) = @_;
    $chapter = lc($chapter);
    $chapter =~ s/\s+(functions|api)\s*$//;
    $chapter = ucfirst($chapter) . " Functions";
    return $chapter;
}

if ($testing) {
    foreach my $f (@functions) {
	dumpfunc($f);
    }
} else {

    # generate texi. I have no idea what texinfo files should look like, I'm
    # just trying to create something that looks a lot like the perlfunc.texi
    # page from the perl-info.5.004 distribution, since that lets you do a
    # C-h f and jump right to the function definition

    # all functions are available in @functions
    # per-chapter are in @{$chapters{$name}}


    # generate one file per chapter heading.
    my(@chapters) = sort(keys(%chapters));
    foreach my $chapter_num (0 .. $#chapters) {
	my $chapter = $chapters[$chapter_num];
	my $listname = chapter_to_nodename($chapter);
	my $texiname = lc($chapter);
	$texiname =~ s/\s+(functions|api)\s*$//;
	my $filename = $texiname . "_funcs.texi";
	$filename =~ s/ /_/g;

	print "Texinfo file name: $filename\n";
	open(TEXI, ">$filename") 
	  or die "couldn't write $filename: $!";

	# first, make a list of function names so we can put appropriate
	# next,prev pointers in
	my(@funcs) = sort {$a->{'name'} cmp $b->{'name'}} @{$chapters{$chapter}};

	# header stuff: make a node that lists all functions in this chapter
	print TEXI "\@node $listname, ";
	if ($chapter_num < $#chapters) {
	    my $next_chapter = chapter_to_nodename($chapters[$chapter_num+1]);
	    print TEXI "$next_chapter, ";
	} else {
	    print TEXI ", ";
	}
	if ($chapter_num > 0) {
	    my $prev_chapter = chapter_to_nodename($chapters[$chapter_num-1]);
	    print TEXI "$prev_chapter, ";
	} else {
	    print TEXI ", ";
	}
	print TEXI "Top\n\n";

	print TEXI "$listname\n\n";

	print TEXI "\@menu\n";
	my $longest = 0;
	foreach my $f (@funcs) {
	    my $len = length($f->{'name'});
	    $longest = $len if $len > $longest;
	}
	$longest += 6;
	$Text::Wrap::columns = 75;
	foreach my $f (@funcs) {
	    my $start = "* $f->{'name'}::  ";
	    $start .= " " x ($longest - length($start));
	    print TEXI expand(wrap($start, " " x $longest, $f->{'purpose'}));
	    print TEXI "\n\n";
	}
	print TEXI "\@end menu\n\n\n";

	# per function
	foreach my $fi (0 .. $#funcs) 
	{
	    $f = $funcs[$fi];
	    print TEXI "\@node $f->{'name'}, ";
	    if ($fi < $#funcs) {
		print TEXI $funcs[$fi+1]->{'name'},", ";
	    } else {
		print TEXI ", ";  # there is no next
	    }
	    if ($fi > 0) {
		print TEXI $funcs[$fi-1]->{'name'},", ";
	    } else {
		print TEXI ", ";  # no prev
	    }
	    print TEXI "$listname\n";

	    print TEXI "\@findex $f->{'name'}\n";

	    print TEXI "$listname\n";
	    print TEXI "\n";

	    print TEXI "\@unnumberedsec NAME\n";
	    print TEXI $f->{'name'},"\n\n";

	    print TEXI "\@unnumberedsec PURPOSE\n";
	    print TEXI $f->{'purpose'},"\n\n";

	    print TEXI "\@unnumberedsec PROTOTYPE\n";
	    print TEXI "\@example\n";
	    my $proto = $f->{'prototype'};
	    print TEXI $proto->[0] , "\n";
	    my $start = $f->{'name'} . "(";
	    my $input_parms = scalar(@{$proto}) - 1;
	    if (!$input_parms) {
		print TEXI $start,"void)\n";
	    } else {
		# get creative with the filling here. Find the longest typename
		my $longest = 0;
		foreach my $i (1 .. $input_parms) {
		    my $len =  length($proto->[$i]->{'type'});
		    $longest = $len if $len > $longest;
		}
		# now put length($start) spaces before all typenames, and
		# pad all typenames out to $longest+2
		print TEXI "$start";
		print TEXI "$proto->[1]->{'type'}";
		print TEXI " " x ($longest+2-length($proto->[1]->{'type'}));
		print TEXI $proto->[1]->{'name'};
		if ($input_parms == 1) {
		    print TEXI ")\n";
		} else {
		    print TEXI ",\n";
		    foreach my $i (2 .. $input_parms) {
			print TEXI " " x length($start);
			print TEXI $proto->[$i]->{'type'};
			print TEXI " " x ($longest+2 - 
					  length($proto->[$i]->{'type'}));
			print TEXI $proto->[$i]->{'name'};
			if ($i == $input_parms) {
			    print TEXI ")\n";
			} else {
			    print TEXI ",\n";
			}
		    }
		}
	    }
	    print TEXI "\@end example\n\n";

	    if ($input_parms) {
		print TEXI "\@unnumberedsec PARAMETERS\n";
		print TEXI "\@example\n";
		# do creative filling here too
		my $longest = 0;
		my $longest_dir = 0;
		foreach my $i (1 .. $input_parms) {
		    my $len = length($proto->[$i]->{'name'});
		    if ($proto->[$i]->{'dir'}) {
			my $len_dir = length($proto->[$i]->{'dir'}) + 1;
			$longest_dir = $len_dir if $len_dir > $longest_dir;
		    }
		    $longest = $len if $len > $longest;
		}
		
		$Text::Wrap::columns = 75 - 5; # the @example indents by 5
		foreach my $i (1 .. $input_parms) {
		    my $parm = $proto->[$i];
		    my $start = "$parm->{'name'}";
		    if ($longest_dir) {
			my $dirlen = 0;
			$dirlen = length($parm->{'dir'}) if $parm->{'dir'};
			$start = " " x ($longest_dir - $dirlen) . " $start";
			$start = $parm->{'dir'} . $start if $parm->{'dir'};
		    }
		    $start .= " " x ($longest+1-length($parm->{'name'}));
		    $start .= ": ";
		    my $text;
		    if ($parm->{'text'}) {
			$text = $parm->{'text'};
		    } else {
			$text = "<no description>";
		    }
		    print TEXI expand(wrap($start, " " x length($start), $text));
		    print TEXI "\n";
		}
		print TEXI "\@end example\n\n";
	    }


	    if ($f->{'result'}) {
		print TEXI "\@unnumberedsec RESULT\n";
		my $result = $f->{'result'};
		$result =~ s/\{/\@\{/g; # texinfo escape for {
		$result =~ s/\}/\@\}/g;
		print TEXI $result,"\n\n";
	    }

	    if ($f->{'comment'}) {
		print TEXI "\@unnumberedsec COMMENTS\n";
		my $comment = $f->{'comment'};
		$comment =~ s/\{/\@\{/g; # texinfo escape for {
		$comment =~ s/\}/\@\}/g;
		print TEXI $comment,"\n\n";
	    }

	    if ($f->{'caveat'}) {
		print TEXI "\@unnumberedsec CAVEAT\n";
		print TEXI $f->{'caveat'},"\n\n";
	    }

	    if ($f->{'seealso'}) {
		print TEXI "\@unnumberedsec SEE ALSO\n";
		my(@refs) = split(/\s+/, $f->{'seealso'});
		foreach my $r (@refs) {
		    print TEXI "\@xref{$r}.\n";
		}
	    }

	    print TEXI "\n";
	}
	# end of texi file stuff
	close(TEXI);
    }
}


# todo:
# CtlShowControl shows up on the Field UI Functions
# chapter names with >2 words, like Field UI Functions: upcase UI
# figure out how to make a pilot-C mode that will take C-h f and actually
#  do the info lookup (perhaps borrow from cperl-mode.el?)
# remove things like "For System Use Only" from the See Also nodes, then
#  turn off --no-pointer-validate
# some parm descriptions span multiple parms. This leads to a sync loss on
#  the parm names causing all the description text to get lumped with the
#  first parm. This happens a lot on functions with paired parms, like x,y
#  or src,dst or foostart,fooend . Turn on $debug_empty_parms to see them.
# guide3.2 has functions with <-> direction indicators *after* the parm name.
#  like SerReceive.
# SlkSendPacket gets headerP direction wrong.

